using System;
using System.Collections;
using System.Drawing;
using System.Diagnostics;
using System.ComponentModel;

namespace CarlosB.PathFind
{
	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	[ToolboxBitmapAttribute(typeof(AStar), "AStar.ico")]
	public class AStar: Component
	{
		private struct NodeData
		{
			public byte TerrainCost;
			public ushort Cycle;
			public bool Opened;
			public double HeuristicCost,
				PathCost,
				TotalCost;
			public int TotalSteps;
			public Point ConnectFrom;
		}

		public Point[] Directions = { new Point(1, 0), new Point(0, 1), new Point(-1, 0), new Point(0, -1), new Point(-1, -1), new Point(-1, 1), new Point(1, -1), new Point(1, 1) };
		private int FWidth = 1,
			FHeight = 1;
		public byte[,] CellCost;
		private float FHeuristicWeight = 1;
		private NodeData[,] FNodes;
		private ushort FCurrentCycle = 0;
		private OpenList FOpenQueue;
		private int FMaxSteps;
		const byte UNTRESPASSABLE_TERRAIN = 255;

		public AStar()
		{
			FOpenQueue = new OpenList(this);
		}

		[DescriptionAttribute("Event fired when needs to estimate the goal cost")]
		public event EstimateGoalCostEventHandler EstimateGoalCost;

		[DescriptionAttribute("Event fired when cell is first visited in order to get the terrain cost")]
		public event CellEventHandler Cell;
		
		[DescriptionAttribute("Height of map")]
		public int Height
		{
			get
			{
				return FHeight;
			}

			set
			{
				if (value < 1)
					FHeight = 1;
				else
					FHeight = value;
			}
		}

		[DescriptionAttribute("Value used to estimate cost to goal, it's multiplied with manhattan distance")]
		public float HeuristicWeight
		{
			get
			{
				return FHeuristicWeight;
			}

			set
			{
				if (value < 1)
					FHeuristicWeight = 1;
				else
					FHeuristicWeight = value;
			}
		}

		[DescriptionAttribute("Maximum steps allowed to get to goal")]
		public int MaxSteps
		{
			get
			{
				return FMaxSteps;
			}

			set
			{
				if (value < 0)
					FMaxSteps = 0;
				else
					FMaxSteps = value;
			}
		}

		[DescriptionAttribute("Width of map")]
		public int Width
		{
			get
			{
				return FWidth;
			}

			set
			{
				if (value < 1)
					FWidth = 1;
				else
					FWidth = value;
			}
		}

		public int FindPath(Point Origin, Point Dest, ref Point[] Steps, out double Cost)
		{
			if ((Cell == null) && 
				((CellCost.GetLength(0) < FWidth) || (CellCost.GetLength(1) < FHeight)))
				throw new Exception("cell array not properly defined");

			int StepsLimit = FMaxSteps;
			if (FMaxSteps <= 0)
				// void maximum step limit
				StepsLimit = Int32.MaxValue;

			// empty OpenQueue
			FOpenQueue.Clear();

			// init internal cell array
			if ((FNodes == null) || (FNodes.GetLength(0) != FWidth) || (FNodes.GetLength(1) != FHeight))
				FNodes = new NodeData[FWidth, FHeight];

			// init cell array cycle identifiers if first cycle
			if (FCurrentCycle == 0)
				for (int x=0; x<FWidth; x++)
					for (int y=0; y<FHeight; y++)
						FNodes[x,y].Cycle = 0;

			FCurrentCycle++;

			// add first position to queue
			NeedCell(Origin, Dest);
			FNodes[Origin.X, Origin.Y].PathCost = 0;
			FNodes[Origin.X, Origin.Y].TotalCost = FNodes[Origin.X, Origin.Y].HeuristicCost;
			FNodes[Origin.X, Origin.Y].TotalSteps = 0;
			FNodes[Origin.X, Origin.Y].Opened = true;
			FNodes[Origin.X, Origin.Y].ConnectFrom = new Point(-1, -1);
			FOpenQueue.Add(Origin);

			while (!FOpenQueue.Empty)
			{
				// get node from open queue
				Point CurrentPos = FOpenQueue.RemoveFirst();
				double CurrentCost = FNodes[CurrentPos.X, CurrentPos.Y].PathCost;
				int CurrentSteps = FNodes[CurrentPos.X, CurrentPos.Y].TotalSteps;

				if ((CurrentPos == Dest) || (CurrentSteps == StepsLimit))
				// reached goal or maximum steps given
				{
					// alloc result path if not done by caller
					if (Steps == null)
						Steps = new Point[CurrentSteps];

					// write out resulting path
					for (int Cnt=CurrentSteps-1; Cnt>=0; Cnt--)
					{
						Steps[Cnt] = CurrentPos;
						CurrentPos = FNodes[CurrentPos.X, CurrentPos.Y].ConnectFrom;
					}

					Cost = CurrentCost;
					return CurrentSteps;
				}
				FNodes[CurrentPos.X, CurrentPos.Y].Opened = false;

				foreach (Point Dir in Directions)
				{
					// calc neighboor position
					Point NewPos = new Point(CurrentPos.X + Dir.X, CurrentPos.Y + Dir.Y);
					if ((NewPos.X < 0) || (NewPos.X >= FWidth) ||
						(NewPos.Y < 0) || (NewPos.Y >= FHeight))
						continue;

					// make sure cell is updated
					NeedCell(NewPos, Dest);

					// check if cell is untrespassable
					if (FNodes[NewPos.X, NewPos.Y].TerrainCost >= UNTRESPASSABLE_TERRAIN)
						continue;

					double absShift = Math.Sqrt(Math.Pow(Dir.X, 2) + Math.Pow(Dir.Y, 2));
					//double absShift = Math.Abs(Dir.X) + Math.Abs(Dir.Y);
					//double absShift = Math.Abs(Dir.X) + Math.Abs(Dir.Y) == 1 ? 1 : 1.5;
					double NewCost = CurrentCost + (FNodes[NewPos.X, NewPos.Y].TerrainCost + 1) * absShift;

					if (NewCost < FNodes[NewPos.X, NewPos.Y].PathCost)
					{
						if (FNodes[NewPos.X, NewPos.Y].Opened)
							// take node from open queue
							FOpenQueue.Remove(NewPos);

						// Add node to open queue
						FNodes[NewPos.X, NewPos.Y].PathCost = NewCost;
						FNodes[NewPos.X, NewPos.Y].TotalCost = NewCost + FNodes[NewPos.X, NewPos.Y].HeuristicCost;
						FNodes[NewPos.X, NewPos.Y].TotalSteps = CurrentSteps + 1;
						FNodes[NewPos.X, NewPos.Y].Opened = true;
						FNodes[NewPos.X, NewPos.Y].ConnectFrom = CurrentPos;
						FOpenQueue.Add(NewPos);
					}
				}
			}

			Cost = 0;
			return -1;
		}

		private void NeedCell(Point Pos, Point Dest)
		{
			// init node if it's first visited this cycle
			if (FNodes[Pos.X, Pos.Y].Cycle != FCurrentCycle)
			{
				CellEventArgs eCell = new CellEventArgs(Pos);
				OnCell(eCell);
				FNodes[Pos.X, Pos.Y].TerrainCost = eCell.Result;
				EstimateGoalCostEventArgs eHeuristic = new EstimateGoalCostEventArgs(Pos, Dest);
				OnEstimateGoalCost(eHeuristic);
				FNodes[Pos.X, Pos.Y].HeuristicCost = eHeuristic.Result;
				FNodes[Pos.X, Pos.Y].Cycle = FCurrentCycle;
				FNodes[Pos.X, Pos.Y].PathCost = Int32.MaxValue;
				FNodes[Pos.X, Pos.Y].Opened = false;
			}
		}
		
		protected virtual void OnCell(CellEventArgs e)
		{
			if (Cell != null)
				Cell(this, e);
			else
				e.Result = CellCost[e.Pos.X, e.Pos.Y];
		}		
		
		protected virtual void OnEstimateGoalCost(EstimateGoalCostEventArgs e)
		{
			if (EstimateGoalCost != null)
				EstimateGoalCost(this, e);
			else
				// by default uses Manhattan distance to estimate cost
				//e.Result = (int)((Math.Abs(e.From.X - e.To.X) + Math.Abs(e.From.Y - e.To.Y)) * FHeuristicWeight);
				e.Result = (int)(Math.Sqrt(Math.Pow(e.From.X - e.To.X, 2) + Math.Pow(e.From.Y - e.To.Y, 2)));
		}		
	
	
	
		private class OpenList: IComparer
			{
				private AStar FOwner;
				private ArrayList FQueue = new ArrayList();
	
				public OpenList(AStar AOwner)
				{
					FOwner = AOwner;
				}

				public void Add(Point Pos)
				{
					int idx = FQueue.BinarySearch(Pos, this);
					FQueue.Insert(-idx-1, Pos);
				}

				public void Clear()
				{
					FQueue.Clear();
				}

				public bool Empty
				{
					get
					{
						return(FQueue.Count == 0);
					}
				}

				public Point RemoveFirst()
				{
					Point ret = (Point)FQueue[0];
					FQueue.RemoveAt(0);
					return ret;
				}

				public void Remove(Point Pos)
				{
					FQueue.RemoveAt(FQueue.BinarySearch(Pos, this));
				}

				public int Compare(object x, object y)
				{
					double ret = (FOwner.FNodes[((Point)x).X, ((Point)x).Y].TotalCost -
						FOwner.FNodes[((Point)y).X, ((Point)y).Y].TotalCost); 

					if (ret == 0)
					{
						ret = ((Point)x).X - ((Point)y).X;
						if (ret == 0)
							ret = ((Point)x).Y - ((Point)y).Y;
					}

					if (ret > 0)
						return 1;
					else if (ret < 0)
						return -1;
					else
						return 0;
				} 
			}
	}
}
